import argparse
import os
from datetime import datetime
from typing import Iterable

try:
    import qai.scraper
except ImportError:
    import os
    import sys

    sys.path.append(os.getcwd())

from qai.scraper.scrapers.stealthscraper import StealthScraper
from qai.scraper.scrapers.scraper import Scraper

default_exclude_regexes = set()
default_exclude_regexes.add(":.*@")
default_exclude_regexes.add("/ja-jp/")


def prep_scrap(
    site_name: str,
    urls: Iterable[str],
    base_dir: str = None,
    dest_dir: str = None,
    exclude_urls: Iterable[str] = None,
    exclude_regexs: Iterable[str] = None,    
    medata_loc: str = None,
    max_depth: int = 4,
    stealth=False,
):
    if exclude_urls is None:
        exclude_urls = set()
    if exclude_regexs is None:
        exclude_regexs = set()
    if base_dir is None:
        now_str = datetime.utcnow().strftime("%Y%m%d-%H%M%S")
        base_dir = f"~/data/websites/{site_name}/{now_str}"
    if dest_dir is None:
        dest_dir = os.path.expanduser(f"{base_dir}/raw")
    if medata_loc is None:
        medata_loc = os.path.expanduser(f"{base_dir}/metadata.csv")
    os.makedirs(base_dir, exist_ok=True)
    
    s = StealthScraper() if stealth else Scraper()
    
    exclude_regexs.add(":.*@")
    exclude_regexs.add("/ja-jp/")
    s.scrape(
        urls,
        dest_dir=dest_dir,
        metadata_loc=medata_loc,
        exclude_urls=exclude_urls,
        exclude_regexs=exclude_regexs,
        max_depth=max_depth,
    )


if __name__ == "__main__":
    ap = argparse.ArgumentParser()
    ap.add_argument("--site-name", type=str, required=True)
    ap.add_argument("--url", type=str, required=True)
    ap.add_argument("--max-depth", type=int, default=-1)
    ap.add_argument("--stealth", action="store_true", default=False)
    # import sys
    # sys.argv = ["", "--site-name", "cognite", "--url", "https://www.cognite.commailto:MRMartin@rockwellautomation.com/en-us/products/details.800T-2TAPW.html", "--max-depth", "-1"]
    args = ap.parse_args()
    extract(site_name=args.site_name, urls=[args.url], max_depth=args.max_depth, stealth=args.stealth)



from collections import defaultdict
from dataclasses import dataclass
from typing import Any

from bs4 import BeautifulSoup, Tag

from .processors import Processor


@dataclass
class TagKeeper(Processor):
    keep_parents = True
    keep_children = True
    html: bool = True
    body: bool = True
    br: bool = True
    header: bool = False
    footer: bool = False
    javascript: bool = False
    css: bool = False
    style: bool = False
    images: bool = False
    empty: bool = False
    meta: bool = False
    document: bool = False
    buttons: bool = False
    h1: bool = False
    h2: bool = False
    h3: bool = False
    ol: bool = False
    ul: bool = False
    li: bool = False
    div: bool = False
    classes: list[str] = None
    name: str = None

    def process(
        self,
        soup: BeautifulSoup,
        instance_info: dict[str, Any] = None,
    ) -> BeautifulSoup:
        if not self.classes:
            self.classes = []
        locked = set()
        parent_map = defaultdict(set)
        to_keep = []
        if self.html:
            locked.add("html")
        if self.body:
            locked.add("body")
        if self.br:
            locked.add("br")
        if self.h1:
            to_keep += ["h1"]
        if self.h2:
            to_keep += ["h2"]
        if self.h3:
            to_keep += ["h3"]
        if self.header:
            to_keep += ["header", "head", "page-header"]
            self.classes += ["header", "head", "page-header"]
        if self.footer:
            to_keep += ["footer", "foot", "page-footer"]
        if self.javascript:
            to_keep += ["script", "noscript"]
        if self.css:
            to_keep += ["css"]
        if self.images:
            to_keep += ["img"]
        if self.style:
            to_keep += ["style"]
        if self.meta:
            to_keep += ["meta"]
        if self.document:
            to_keep += ["[document]"]
        if self.buttons:
            to_keep += ["button"]
        if self.ol:
            to_keep += ["ol"]
        if self.ul:
            to_keep += ["ul"]
        if self.li:
            to_keep += ["li"]
        if self.div:
            to_keep += ["div"]

        to_keep = set(to_keep)
        to_keep_classes = set(self.classes)

        # Find all tags, then filter out the irrelevant ones
        remove = set()
        keep = set()

        def is_a_parent(e: Tag, tocheck: Tag) -> bool:
            parents = parent_map.get(e, set())
            if tocheck in parents:
                return True
            for p in parents:
                parents.update(is_a_parent(p))
            return tocheck in parents

        for e in soup.find_all():
            if e in keep:
                continue

            should_keep = False
            if e.name in to_keep:
                should_keep = True
            else:
                classes = e.get("class")
                if classes:
                    classes = set(classes)
                    if classes.intersection(to_keep_classes):
                        should_keep = True

            ## keep all tags that are children of the kept tag
            if should_keep and self.keep_children:
                keep.update(e.find_all())
            if should_keep and self.keep_parents:
                keep.update(e.find_parents())

            if e.name in locked:
                should_keep = True
            if should_keep:
                keep.add(e)
            else:
                remove.add(e)
        remove = remove - keep
        for e in remove:
            if e and not e.decomposed:
                e.decompose()

        return soup



from dataclasses import dataclass
from typing import List, Set

from bs4 import BeautifulSoup

from .processors import Processor
import re

def prettify_tag(html, tag):
    reg_tag = re.compile(f'\s*<{tag}([^>]*)>\s*')
    html = reg_tag.sub(f'<{tag}\\1>', html)
    html = re.sub(f'\s*</{tag}>\s*',f'</{tag}>', html)
    return html

@dataclass
class TagDenester(Processor):
    div: bool = True
    span: bool = True
    spans_become_spaces = True
    all_others: bool = True

    def process(
        self,
        soup: BeautifulSoup,
    ) -> BeautifulSoup:
        for e in soup.find_all():
            if self.div:
                if e.name == "div":
                    if e.parent.name == "div":
                        e.unwrap()
            if self.span:
                if e.name == "span":
                    e.unwrap()
            if self.all_others:
                if e.name not in ["div", "span"]:
                    if e.parent.name == e.name:
                        e.unwrap()
            # if self.span and e.name == "span":
            #     pe = e.parent
            #     # if self.spans_become_spaces:
            #         # pe.string += " " + e.get_text().strip()
            #     e.unwrap()
        return soup



from dataclasses import dataclass

import pandas as pd
from bs4 import BeautifulSoup

from qai.scraper.filters.meta import InstanceInfo
from qai.scraper.processors.processors import Processor
from qai.scraper.utils.bs_utils import html_strip



@dataclass
class DivTable(Processor):

    # container: str = None
    table: str = None
    columns: list[str] = None
    trs: list[str] = None
    column_headers: list[str] = None
    
    # has_header: bool = Truet

    def process(
        self, soup: BeautifulSoup, instance_info: InstanceInfo = None, **kwargs
    ) -> BeautifulSoup:
        table = soup.find("div", {"name": self.table})
        if not table:
            return soup
        data = {}
        headers = []
        column_headers = self.column_headers if self.column_headers else [None] * len(self.columns)
        for column, row, header in zip(self.columns, self.trs, column_headers):
            vals = []
            ## Find all divs with the name <container>
            container = table.find("div", {"name": column})
            if header:
                h = container.find("div", {"name": header}).extract()
                txt = html_strip(h.text)
                headers.append(txt)
            else:
                headers.append(column)

            ## Find all divs with the name <tr>
            trs = container.find_all("div", {"name": row})
            print(len(trs))
            # print(trs[0])
            for tr in trs:
                # print(tr)
                txt = html_strip(tr.text)
                vals.append(txt)
            data[column] = vals
        if not headers:
            headers = list(data.keys())
        df = pd.DataFrame(data=data)
        df.columns = headers
        print(df.to_html(index=False))
        ## replace the table with the new table
        table.replace_with(BeautifulSoup(df.to_html(index=False), "html.parser"))
        return soup



import os
import tomllib as toml
from typing import Iterable

from qai.core import Meta
from qai.scraper.filters.filter import MultiFilter
from qai.scraper.scrapers.scraper import Scraper

filter_cfg = """
[[pipeline]]
in_folder = "raw"
out_folder = "simplified"
categories = ["general"]
pipeline = ["Remove", "Simplify", "HtmlWriter", "MarkdownWriter"]

[[processors]]
name = "Remove"
header = true
footer = true
javascript = true
css = true
style = true
images = true
empty = true
meta = true
document = true
buttons = true
p_in_a = true
classes = ["home-header", "home-footer", "main-header", "main-footer"]

[[processors]]
name = "Simplify"
unwrap_a = false
sanitize = false

[[processors]]
name = "HtmlWriter"
out_folder = "simplified_html"

[[processors]]
name = "MarkdownWriter"
out_folder = "simplified_md"
"""


class ScraperFilter:
    def __init__(
        self,
        meta: Meta = None,
        webdir: str = None,
        within_seconds: int = 0,
        resume: bool = False,
        scraper: Scraper = None,
        filter: MultiFilter = None,
    ):
        self.scraper = scraper
        self.filter = filter
        self.meta = meta
        if self.meta is None:
            if not webdir:
                raise ValueError("Either meta or webdir must be provided.")
            if resume and os.path.exists(webdir):
                print(f"Directory {webdir} already exists. Using existing data.")
                meta = Meta.from_most_recent(webdir)
            else:
                print(f"Creating new meta data in {webdir}.")
                meta = Meta.create_most_recent(webdir)
        if self.filter is None:
            filter_cfgs = [
                toml.loads(filter_cfg),
            ]
            self.filter = MultiFilter(filter_cfgs, meta=meta)
        if self.scraper is None:
            self.scraper = Scraper()

    def scrape_and_filter(
        self,
        urls: Iterable[str],
        dest_subfolder: str = "raw",
        exclude_urls: Iterable[str] = None,
        exclude_regexs: Iterable[str] = None,
        max_depth: int = 4,
        subfolder="raw",
    ):
        dest_dir = self.meta.get_group(dest_subfolder, create=True)

        self.scraper.scrape(
            urls,
            dest_dir=dest_dir,
            exclude_urls=exclude_urls,
            exclude_regexs=exclude_regexs,
            max_depth=max_depth,
            group=subfolder,
            meta=self.meta,
        )
        self.filter.process_directory(self.meta.get_group(dest_subfolder))



from dataclasses import dataclass
from typing import Any

from bs4 import BeautifulSoup

from .processors import Processor


@dataclass
class TagRemover(Processor):
    header: bool = False
    footer: bool = False
    javascript: bool = False
    css: bool = False
    style: bool = False
    images: bool = False
    img: bool = False
    svg: bool = False
    empty: bool = False
    meta: bool = False
    document: bool = False
    buttons: bool = False
    ol: bool = False
    ul: bool = False
    a: bool = False
    b: bool = False
    p_in_a: bool = False
    classes: list[str] = None
    tags: list[str] = None

    def process(
        self, soup: BeautifulSoup, instance_info: dict[str, Any] = None, **kwargs
    ) -> BeautifulSoup:
        to_remove_classes = [] if not self.classes else self.classes
        to_remove = []
        if self.header:
            to_remove += ["header", "head", "page-header"]
            to_remove_classes += ["header", "head", "page-header"]
        if self.footer:
            to_remove += ["footer", "foot", "page-footer"]
        if self.javascript:
            to_remove += ["script", "noscript"]
        if self.css:
            to_remove += ["css"]
        if self.images:
            to_remove += ["img", "svg"]
        if self.img:
            to_remove += ["img"]
        if self.svg:
            to_remove += ["svg"]
        if self.style:
            to_remove += ["style"]
        if self.meta:
            to_remove += ["meta"]
        if self.document:
            to_remove += ["[document]"]
        if self.buttons:
            to_remove += ["button"]
        if self.b:
            to_remove += ["b"]
        if self.a:
            to_remove += ["a"]
        if self.ol:
            to_remove += ["ol"]
        if self.ul:
            to_remove += ["ul"]
        if self.tags:
            to_remove += self.tags
        if to_remove:
            for e in soup.find_all(to_remove):
                e.decompose()
        ## remove p tags inside a tags
        if self.p_in_a:
            for e in soup.find_all("a"):
                for p in e.find_all("p"):
                    p.unwrap()
        to_remove_classes = set(to_remove_classes)
        # print(soup.prettify())
        for c in to_remove_classes:
            for e in soup.find_all(class_=c):
                e.decompose()
        if self.empty:
            locked = set(["html", "body", "br"])
            for e in soup.find_all():
                if not e.get_text(strip=True).strip() and e.name not in locked:
                    e.decompose()

        return soup


@dataclass
class CleanHtml(TagRemover):
    javascript: bool = True
    css: bool = True
    style: bool = True
    images: bool = True



import os
from dataclasses import dataclass
from typing import Any
from qai.core import MetaFile, MetaDir

from qai.scraper.filters.meta import InstanceInfo

# from qai.scraper.scrapers.scraper import WebObj

from .processors import Processor


@dataclass
class Writer(Processor):
    file_name: str = None
    out_folder: str = None

    def __post_init__(self):
        if self.meta is None:
            raise ValueError(
                "meta must be specified for Writer as there must be some place to write to"
            )

    def _append_and_modify_filename(
        self, extension: str = None, instance_info: InstanceInfo = None
    ) -> MetaDir:
        if self.out_folder is not None:
            webobj: MetaFile = instance_info.web_file
            meta = instance_info._meta

            group = meta.get_dir(self.out_folder, create=True)
            if self.file_name is None:
                self.file_name = group.path / webobj.path.name
            dn = os.path.dirname(self.file_name)
            os.makedirs(dn, exist_ok=True)
            if extension is not None:
                fn, __ = os.path.splitext(self.file_name)
                if extension.startswith("."):
                    extension = extension[1:]
                self.file_name = f"{fn}.{extension}"
            # print("Writing to", self.file_name, self.out_folder)
            # mf = MetaFile(
            #     path=self.file_name,metadata={"instance_info": instance_info}
            # )
            # mf.add_origin(webobj)

            # group.add_file(mf)
            # self.meta.add_file_meta(
            #     group=self.out_folder,
            #     values={
            #         "uri": webobj.uri,
            #         "file_name": os.path.basename(self.file_name),
            #         "source_uri": f"{{group:{webobj._group}}}/{webobj.uri}",
            #     },
            #     # write_to_file="w",
            # )
            # self.meta.save(overwrite=True)
            return group

    # def process(self, soup: BeautifulSoup, **kwargs) -> BeautifulSoup:



from dataclasses import dataclass
from typing import Any, Optional

from qai.core import Meta, MetaPath, MetaFile
# from qai.scraper.scrapers.meta import WebFile
from pydantic import BaseModel

class DirectoryInfo(BaseModel):
    in_folder: Optional[str] = None
    _meta: Optional[Meta] = None

class InstanceInfo(BaseModel):
    web_file: MetaFile = None
    _meta: Meta = None
    config: dict[str, Any] = None
    config_index: int = None
    directory_info: DirectoryInfo = None


import os
from dataclasses import dataclass
from typing import Any

import ftfy
from bs4 import BeautifulSoup

from qai.scraper.filters.meta import InstanceInfo
from qai.scraper.processors.writer import Writer
from qai.core import MetaFile, MetaDir

@dataclass
class HTMLWriter(Writer):
    file_name: str = None
    prettify: bool = True

    def process(
        self, soup: BeautifulSoup, instance_info: InstanceInfo = None, **kwargs
    ) -> BeautifulSoup:
        group = self._append_and_modify_filename("html", instance_info=instance_info)
        if self.file_name is not None:
            with open(self.file_name, "w") as fp:
                html_s = ftfy.fix_text(str(soup.prettify()))
                soup = self.make_soup(html_s)
                html_s = str(soup.prettify())
                fp.write(html_s)
                if self.meta:
                    mf = MetaFile(path=self.file_name, metadata={"instance_info": instance_info})
                    if group:
                        group.add_file(mf)
                    else:
                        self.meta.add_file(mf)
        return soup



import re
from dataclasses import dataclass
from typing import List

import ftfy
from bs4 import BeautifulSoup, NavigableString, Tag

from qai.scraper.processors.tag_denester import TagDenester
from qai.scraper.utils.bs_utils import sanitize as sanitize_html_str

from .processors import Processor


@dataclass
class HTMLSimplifier(Processor):
    file_name: str = None
    sanitize: bool = True
    denest: bool = True
    simplify_p: bool = True
    empty: bool = True
    unwrap_a: bool = True
    unwrap_b: bool = True
    unwrap_i: bool = True
    unwrap_em: bool = True
    unwrap_strong: bool = True
    unwrap_it: bool = True
    unwrap_img: bool = False
    value_from_attributes: list[str] = None

    def process(self, soup: BeautifulSoup, file_name: str = None, **kwargs) -> BeautifulSoup:
        if file_name is None:
            file_name = self.file_name
        html_str = ftfy.fix_text(str(soup))
        soup = self.make_soup(html_str)

        if self.sanitize:
            html_str = sanitize_html_str(str(soup))
            soup = self.make_soup(html_str)
        ## change values if they are dynamically set in attributes
        if self.value_from_attributes:
            attrs = set(self.value_from_attributes)
            for tag in soup.find_all(lambda t: len(t.attrs) and attrs.intersection(t.attrs.keys())):
                for a in attrs.intersection(tag.attrs.keys()):
                    tag.string = tag.attrs[a]
        ## Unwrap tags
        unwrap_tags = set()
        if self.unwrap_img:
            unwrap_tags.add("img")
        if self.unwrap_a:
            unwrap_tags.add("a")
        if self.unwrap_b:
            unwrap_tags.add("b")
        if self.unwrap_i:
            unwrap_tags.add("i")
        if self.unwrap_em:
            unwrap_tags.add("em")
        if self.unwrap_strong:
            unwrap_tags.add("strong")
        if self.unwrap_it:
            unwrap_tags.add("it")
        if self.denest:
            soup = TagDenester().process(soup)

        for tag in soup.find_all(unwrap_tags):
            tag.unwrap()

        if self.empty:
            locked = set(["html", "body", "br"])
            for e in soup.find_all():
                if not e.get_text(strip=True).strip() and e.name not in locked:
                    e.decompose()

        ## Simplify text
        e: Tag
        for e in soup.find_all():
            txt = e.find(string=True, recursive=False)
            if not txt:
                continue
            size = len(e.contents)
            for i, c in enumerate(e.contents):
                if isinstance(c, NavigableString) and c.parent == e:
                    modify = True
                    if i < size - 1:
                        nxt = e.contents[i + 1]
                        if isinstance(nxt, NavigableString):
                            nxt.replace_with(str(c) + " " + str(nxt))
                            c.replace_with("")
                            modify = False
                    if modify:
                        s = str(c).replace("\n", " ")
                        s = s.replace("\t", " ")
                        s = re.sub(r"\s+", " ", s)
                        ## Remove spaces before punctuation
                        s = re.sub(r"\s+[.]", ".", s)
                        s = re.sub(r"\s+[?]", "?", s)
                        s = re.sub(r"\s+[!]", "!", s)
                        s = re.sub(r"\s+[;]", ";", s)
                        s = re.sub(r"\s+[,]", ",", s)
                        ## Correctly format percentages (50%)
                        s = re.sub(r"(\d+)\s+[%]", r"\1%", s)
                        ## Times symbol (3x speed, or 4x4)
                        s = re.sub(r"(\d+)\s+x(\s)", r"\1x\2", s)
                        s = re.sub(r"(\d+)x\s+(\d+)", r"\1x\2", s)
                        c.replace_with(s)

        if file_name:
            with open(file_name, "w") as fp:
                html_s = str(soup.prettify())
                fp.write(html_s)
        return soup



import re
from dataclasses import dataclass
from typing import Any

import ftfy
from bs4 import BeautifulSoup, NavigableString, Tag

from qai.scraper.processors.tag_denester import TagDenester
from qai.scraper.utils.bs_utils import sanitize as sanitize_html_str

from .processors import Processor


@dataclass
class ValueConverter(Processor):
    value_from_attributes: list[str] = None

    def process(
        self, soup: BeautifulSoup, instance_info: dict[str, Any] = None
    ) -> BeautifulSoup:
        ## change values if they are dynamically set in attributes
        if self.value_from_attributes:
            attrs = set(self.value_from_attributes)
            for tag in soup.find_all(lambda t: len(t.attrs) and attrs.intersection(t.attrs.keys())):
                for a in attrs.intersection(tag.attrs.keys()):
                    tag.string = tag.attrs[a]
        return soup



import logging
import os
import random
import re
import time
import traceback
from dataclasses import dataclass, field
from datetime import datetime
from logging import getLogger
from typing import Any, Set, Union, Optional
from urllib.parse import urlsplit

# import chromedriver_autoinstaller
import undetected_chromedriver as uc
from bs4 import BeautifulSoup
from qai.core import Meta, MetaBase, MetaDir, MetaFile, MetaPath

from qai.scraper.scrapers.meta import ScrapeMeta
from qai.scraper.utils.bs_utils import get_links, url_with_path
from selenium.webdriver.chrome.webdriver import WebDriver

log = getLogger(__name__)


# chromedriver_autoinstaller.install()
## Can check at https://useragentstring.com/, https://useragentstring.com/pages/Chrome/

user_agents = [
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.79 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.84 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.77 Safari/537.36",
]



class ScrapeSession(MetaBase):
    start_urls: list[str] = field(default_factory=list)
    dest_dir: str = None
    searched_urls: list[str] = field(default_factory=list)
    group: str = None
    meta: Meta = None
    submeta: dict[str, Any] = None


class Scraper:
    count = 0

    def __init__(self, options=None, driver=None, offline=False):
        self.driver = self._new_driver(quit=False, options=options, driver=driver)
        self.scrape_sessions = []
        self.offline = offline
        self.time_between_requests = 1
        self.time_jitter_between_requests = 1
        self.last_request_time = time.time()

        if self.driver is None:
            raise ValueError("Failed to create driver")

    def _new_driver(
        self,
        quit=True,
        options=None,
        driver=None,
        random_user_agent=False,
        version_main=None,  ## Specific version of chrome to use
    ) -> WebDriver:
        if quit:
            self.quit()
        if options is None:
            options = uc.ChromeOptions()
            options.add_argument("--disable-blink-features=AutomationControlled")
            options.add_argument("--headless=True")
        if driver is None:
            try:
                driver = uc.Chrome(use_subprocess=True, options=options, version_main=version_main)
            except Exception as e:
                log.error(f"Failed to create driver, uc.version={uc.__version__}")
                log.error(e)

                return

            if random_user_agent:
                agent = random.choice(user_agents)
            else:
                agent = user_agents[0]
            driver.execute_cdp_cmd(
                "Network.setUserAgentOverride",
                {"userAgent": agent},
            )

        return driver

    def _scrape(
        self,
        url: str,
        dest_dir: str,
        current_depth: int = 0,
        max_depth: int = -1,
        overwrite: bool = False,
        exclude_urls: Set[str] = None,
        exclude_regexs: Set[str] = None,
        scrape_session: ScrapeSession = None,
        exclude_parameters: bool = True,
        submeta: MetaDir = None,
        meta: Meta = None,
    ) -> ScrapeSession:
        if exclude_regexs is None:
            exclude_regexs = set()
        if exclude_urls is None:
            exclude_urls = set()
        # urimeta = scrape_session.meta.get_uri_meta(scrape_session.group)

        if exclude_parameters:
            url = url_with_path(url)
        if exclude_urls is None:
            exclude_urls = set()
        if not isinstance(exclude_urls, set):
            exclude_urls = set(exclude_urls)
        if exclude_regexs is None:
            exclude_regexs = set()
        if not isinstance(exclude_regexs, set):
            exclude_regexs = set(exclude_regexs)
        # if scrape_session.meta is None:
        #     raise ValueError("ScrapeSession.meta is None")
        id = Scraper.count
        Scraper.count += 1
        if url in exclude_urls:
            # print("  " * current_depth, " --- excluded", url)
            return scrape_session
        for rgx in exclude_regexs:
            if re.search(rgx, url):
                print("  " * current_depth, " --- excluded based on match", url, rgx)
                return scrape_session
        # print("---- ", hex(id(exclude_urls)), exclude_urls)
        print(
            "  " * current_depth,
            # f"{current_depth}:{max_depth} Scraping: '{url}', url_with_path='{url_with_path(url)}'",
            f"{current_depth}:{max_depth} Scraping: '{url}'",
        )
        exclude_urls.add(url)
        hashed_name = self.url_to_local_name(url)

        local_file = f"{dest_dir}/{hashed_name}"
        # file_dest_dir = os.path.dirname(local_file)
        # dest_dir = os.path.dirname(local_file)
        os.makedirs(os.path.dirname(local_file), exist_ok=True)
        try:
            print("dest_dir", dest_dir, "hashed_name", hashed_name, "local_file", local_file)
            save_path = self.get_and_save(url, local_file, overwrite=overwrite)
            if save_path:
                # filemeta = MetaFile(path=local_file)
                fmeta = {
                    "hashed_name": hashed_name,
                    "source_uri":url,
                    "is_start_url": url in scrape_session.start_urls,
                    "depth": current_depth,
                    "type": "html_scrape",
                    "category":"Unknown"
                }
                filemeta = MetaFile(
                    path=local_file,
                    metadata=fmeta
                )
                submeta.add_file(filemeta)
        except Exception as e:
            import sys

            with open(f"errors.txt", "a") as fp:
                fp.write(url + "\n")
            traceback.print_exc(file=sys.stdout)
            # print(e)
            logging.error(f"Failed to scrape {url}")
            return scrape_session
        print( "  " * current_depth, " ---- ")
        
        if save_path:
            meta.save(overwrite=True)


        if max_depth and max_depth != -1 and current_depth + 1 >= max_depth:
            return scrape_session

        ## Recursively scrape
        links = get_links(
            url, local_file, exclude_parameters=exclude_parameters, depth=current_depth
        )
        for link in links:
            if exclude_urls and link in exclude_urls:
                continue
            ## TODO Need to add both specific link and base link
            self._scrape(
                url=link,
                dest_dir=dest_dir,
                current_depth=current_depth + 1,
                max_depth=max_depth,
                overwrite=overwrite,
                exclude_urls=exclude_urls,
                exclude_regexs=exclude_regexs,
                scrape_session=scrape_session,
                submeta=submeta,
                meta=meta
            )
        return scrape_session

    def scrape(
        self,
        urls: Union[str, list[str]],
        dest_dir: str | MetaDir,
        current_depth: int = 0,
        max_depth: int = 4,
        overwrite: bool = False,
        exclude_urls: Optional[set[str]] = None,
        exclude_regexs: Optional[set[str]] = None,
        scrape_session: Optional[ScrapeSession] = None,
        group: str = "raw",
        meta: Optional[Meta] = None,
        exclude_parameters: bool = True,
    ) -> ScrapeSession:
        if isinstance(urls, str):
            urls = [urls]
        urls = [url_with_path(u) for u in urls]
        if exclude_regexs is None:
            exclude_regexs = set()
        if exclude_urls is None:
            exclude_urls = set()
        if isinstance(dest_dir, MetaDir):
            dest_dir = str(dest_dir.path)
        if scrape_session is None:
            if meta is None:
                # meta = Meta.from_directory(os.path.dirname(dest_dir), create=True)
                meta = Meta.from_dir(os.path.dirname(dest_dir), create=True)
            scrape_session = ScrapeSession(start_urls=urls, dest_dir=dest_dir, group=group )
            self.scrape_sessions.append(scrape_session)
            meta.add_metadata("scrape_session", self.scrape_sessions)
            meta.save(overwrite=True)
        if meta is None:
            raise ValueError("Meta is None")
        # if scrape_session.meta is None:
        #     if meta:
        #         scrape_session.meta = meta
        #     else:
        #         raise ValueError("ScrapeSession.meta is None")

        # submeta = meta.get_group_meta(group)
        submeta = meta.get_dir(group, create=True)
        # if "scrape_session" not in submeta.metadata:
        #     submeta.metadata["scrape_session"] = []
        scrape_meta = ScrapeMeta(
            subfolder=scrape_session.group,
            start_urls=urls,
            dest_dir=dest_dir,
            exclude_regexs=list(exclude_regexs),
            exclude_urls=list(exclude_urls),
            exclude_parameters=exclude_parameters,
            max_depth=max_depth,
        )
        # submeta.metadata["scrape_session"].append(scrape_meta.model_dump(mode="json")) ##TODO
        # scrape_session.submeta = submeta
        submeta.add_metadata("scrape_session", scrape_meta)

        for url in urls:
            self._scrape(
                url,
                dest_dir=dest_dir,
                current_depth=current_depth,
                max_depth=max_depth,
                overwrite=overwrite,
                exclude_urls=exclude_urls,
                exclude_regexs=exclude_regexs,
                scrape_session=scrape_session,
                submeta=submeta,
                meta=meta
            )
        meta.save(overwrite=True)
        return scrape_session

    @staticmethod
    def base_url(url):
        split_url = urlsplit(url)
        return f"{split_url.scheme}://{split_url.netloc}"

    def get(self, url):
        d = self.driver.get(url)
        self.last_request_time = time.time()
        return d

    @staticmethod
    def url_to_local_name(url: str) -> str:
        """Create a semi human readable name for the url"""

        base_url = url_with_path(url)
        base_url = re.sub(r"https?(://)?", "", base_url)

        base_url = base_url.strip("/")
        return f"{base_url}/index.html"

    def _wait_if_needed(self):
        random_wait = random.randint(
            self.time_between_requests,
            self.time_between_requests + self.time_jitter_between_requests,
        )
        dif = (self.last_request_time + random_wait) - time.time()
        if dif > 0:
            time.sleep(dif)

    def get_and_save(self, url, dest_path, overwrite=False) -> str | None:
        save_name = self.url_to_local_name(url)

        logging.debug(f"Scraper.get_and_save:{url}, {dest_path}")
        if not overwrite and os.path.exists(dest_path):
            logging.debug(f"Scraper.get_and_save: skipping {url}, {save_name}")
            return None
        self.get(url)
        text = str(BeautifulSoup(self.driver.page_source, "lxml"))
        if "Security check" in text and "RayID" in text:
            logging.error(f"Security Check Failed:{url}")
            return None
        if "Request unsuccessful. Incapsula incident ID" in text:
            logging.error(f"Incapsula incident ID:{url}")
            return None
        with open(dest_path, "w") as fp:
            fp.write(self.driver.page_source)
        return dest_path

    def quit(self):
        self.driver.quit()



from dataclasses import dataclass

import ftfy
from bs4 import BeautifulSoup
import re
from .processors import Processor
from markdownify import markdownify as md

@dataclass
class TextWriter(Processor):
    file_name: str = None
    body_width: int = None

    def process(self, soup: BeautifulSoup, typ:int=True) -> BeautifulSoup:
        with open(self.file_name, "w") as fp:
            html = ftfy.fix_text(str(soup.prettify()))
            if typ==0:
                new_str = md(html)
            elif typ==1:
                import html2text
                h = html2text.HTML2Text(bodywidth=self.body_width)
                new_str = h.handle(html)
            else:
                from markdown2 import Markdown
                markdowner = Markdown()

            # new_str = md(html, strip=[])
        
            # new_str = re.sub(r'\n\s*\n', r'\n\n', new_str, flags=re.M)
            # new_str = re.sub(r'\n+', '\n', new_str).strip()

            fp.write(new_str)

        return soup



# Basic Config
filter_cfg = """
[[pipeline]]
in_folder = "raw"
out_folder = "simplified"
categories = ["general"]
pipeline = ["Remove", "Simplify", "HtmlWriter", "MarkdownWriter"]

[[processors]]
name = "Remove"
header = true
footer = true
javascript = true
css = true
style = true
images = true
empty = true
meta = true
document = true
buttons = true
p_in_a = true
classes = ["home-header", "home-footer", "main-header", "main-footer"]

[[processors]]
name = "Simplify"
unwrap_a = false
sanitize = false

[[processors]]
name = "HtmlWriter"
out_folder = "simplified_html"

[[processors]]
name = "MarkdownWriter"
out_folder = "simplified_md"
"""



import os
import re
from dataclasses import dataclass, fields
from typing import Any, ClassVar, List, Tuple, Dict

import ftfy
import pandas as pd
import validators
from bs4 import BeautifulSoup

from qai.scraper.scrapers.scraper import Scraper
from qai.scraper.utils.bs_utils import base_url, clean_soup
from qai.scraper.utils.bs_utils import make_soup as bs_util_make_soup
from qai.scraper.utils.bs_utils import url_with_path
from qai.core import Meta


@dataclass(init=False)
class Processor:
    name: str = None
    meta: Meta = None
    default_parser: str = "lxml"

    def __post_init__(self):
        if self.name is None:
            self.name = self.__class__.__name__

    def __call__(
        self, soup: BeautifulSoup, instance_info: dict[str, Any] = None, **kwargs
    ) -> BeautifulSoup:
        return self.process(soup, instance_info=instance_info, **kwargs)

    @classmethod
    def acceptable_fields(cls) -> set[str]:
        return set([f.name for f in fields(cls)])

    def make_soup(self, html_str_or_path_or_fp: str, default_parser: str = "lxml") -> BeautifulSoup:
        return bs_util_make_soup(html_str_or_path_or_fp, default_parser=default_parser)

    def process(
        self, soup: BeautifulSoup, instance_info: dict[str, Any] = None, **kwargs
    ) -> BeautifulSoup:
        ...



import argparse
import os
from datetime import datetime
from typing import Iterable, Optional

try:
    import qai.scraper
except ImportError:
    import os
    import sys

    sys.path.append(os.getcwd())

import os
import tomllib as toml

import pandas as pd
from qai.core import Meta

from qai.scraper import Scraper
from qai.scraper.filters.basic_filter import filter_cfg
from qai.scraper.filters.filter import MultiFilter
from qai.scraper.scrapers.scraper import Scraper
from qai.scraper.scrapers.scraper_factory import get_scraper
from qai.scraper.scrapers.stealthscraper import StealthScraper


def scrape_filter(
    company,
    urls: str | list[str],
    web_dir: str,
    max_depth=1,
    filter_config=None,
    skip_if_existing: bool = True,
):
    if isinstance(urls, str):
        urls = [urls]
    if filter_config is None:
        filter_config = filter_cfg
    if skip_if_existing and os.path.exists(web_dir):
        print(f"Skipping {company} as it already exists")
        return
    # webdir = os.path.expanduser(f"~/data/websites4/{company}")
    meta = Meta.from_dir(web_dir, create=True, set_save_location=True)
    dest_dir = meta.get_dir("raw", create=True)
    ## prepend http if not present
    urls = [url if url.startswith("http") else f"https://{url}" for url in urls]
    scraper = Scraper()
    scraper.scrape(urls, dest_dir, max_depth=max_depth, meta=meta)
    filter = MultiFilter([filter_config], meta=meta)
    filter.process_directory(in_group="raw")
    meta.save(overwrite=True)



from typing import Any

from qai.core import Meta

from qai.scraper.processors.div_table import DivTable
from qai.scraper.processors.html_simplifier import HTMLSimplifier
from qai.scraper.processors.html_writer import HTMLWriter
from qai.scraper.processors.markdown_writer import MarkdownWriter
from qai.scraper.processors.one_deep_merger import OneDeepMerger
from qai.scraper.processors.processors import Processor
from qai.scraper.processors.tag_denester import TagDenester
from qai.scraper.processors.tag_keeper import TagKeeper
from qai.scraper.processors.tag_remover import CleanHtml, TagRemover
from qai.scraper.processors.text_writer import TextWriter
from qai.scraper.processors.value_converter import ValueConverter


def get_processor(name: str = None, config: dict[str, Any] = None, meta:Meta=None) -> Processor:
    if config is None:
        config = {}
    config = config.copy()
    if name is None:
        name = config.get("name")
    if name is None:
        raise ValueError("Processor name must be provided, either as argument or in config")
    name = name.lower()
    cfg_meta = config.pop("meta", meta)
    if meta is None:
        meta = cfg_meta

    if name.startswith("onedeepmerger") or name.startswith("onedeep"):
        cls = OneDeepMerger
    elif name.startswith("cleanhtml") or name.startswith("clean"):
        cls = CleanHtml
    elif name.startswith("htmlwriter") or name.startswith("html"):
        cls = HTMLWriter
    elif name.startswith("tagremover") or name.startswith("remove"):
        cls = TagRemover
    elif name.startswith("tagkeeper") or name.startswith("keep"):
        cls = TagKeeper
    elif name.startswith("htmlsimplifier") or name.startswith("simpli"):
        cls = HTMLSimplifier
    elif name.startswith("tagdenester") or name.startswith("denest"):
        cls = TagDenester
    elif name.startswith("textwriter") or name.startswith("text"):
        cls = TextWriter
    elif name.startswith("markdownwriter") or name.startswith("markdown"):
        cls = MarkdownWriter
    elif name.startswith("valueconverter") or name.startswith("convert"):
        cls = ValueConverter
    elif name.startswith("divtable"):
        cls = DivTable
    else:
        raise NotImplementedError(f"Processor {name} not implemented")
    flds = cls.acceptable_fields()
    cfg = config.copy()
    cfg = {k: v for k, v in cfg.items() if k in flds}
    return cls(meta=meta, **cfg)




import os
from dataclasses import dataclass
from typing import Any, Tuple

import validators
from bs4 import BeautifulSoup

from qai.scraper.filters.meta import InstanceInfo
from qai.scraper.processors.writer import Writer
from qai.scraper.utils.bs_utils import base_url, clean_soup, get_tags_and_links


@dataclass
class OneDeepMerger(Writer):
    file_name: str = None
    in_folder: str = None

    def process(
        self, soup: BeautifulSoup, instance_info: InstanceInfo = None, url : str = None
    ) -> BeautifulSoup:
        
        if url is None:
            if instance_info is not None and instance_info.web_file is not None:
                url = instance_info.web_file.uri
            else:
                raise ValueError("No web_file in instance_info. Base URL must be specified")
        if self.in_folder is None:
            raise ValueError("No in_folder specified for OneDeepMerger")
        b_url = base_url(url)
        # print(f"{original_url} -> b_url=", b_url)
        # print(soup.prettify())
        for a, link in get_tags_and_links(original_url=b_url, file_or_soup=soup ):
            file_name = link.removeprefix(b_url).strip("/")
            # print(a, b_url, link, file_name)
            lname = self.meta.get_file(file_name=file_name, group=self.in_folder)
            bname = os.path.basename(lname)
            lsoup = clean_soup(str(lname))
            cbody = lsoup.find("body").findChildren(recursive=False)
            # body = ''.join(['%s' % x for x in soup.body.contents])
            # if write_to_file:
            #     HTMLWriter().process(f"{write_location}/{file_prefix}_link_{bname}.html", lsoup)
            for body in cbody[::-1]:
                a.insert_after(body)

        return soup



import logging
import os
import random
import time
from hashlib import sha256
from urllib.parse import urlsplit

from bs4 import BeautifulSoup

from .scraper import Scraper

log = logging.getLogger(__name__)

# https://stackoverflow.com/questions/41721734/take-screenshot-of-full-page-with-selenium-python-with-chromedriver/52572919#52572919
class VisionScraper(Scraper):

    def get_and_save(self, url, dest_path, overwrite=False) -> str | None:
        save_name = self.url_to_local_name(url)

        log.debug(f"Scraper.get_and_save:{url}, {dest_path}")
        if not overwrite and os.path.exists(dest_path):
            log.debug(f"Scraper.get_and_save: skipping {url}, {save_name}")
            return None
        self.get(url)
        text = str(BeautifulSoup(self.driver.page_source, "lxml"))
        if "Security check" in text and "RayID" in text:
            log.error(f"Security Check Failed:{url}")
            return None
        if "Request unsuccessful. Incapsula incident ID" in text:
            log.error(f"Incapsula incident ID:{url}")
            return None
        with open(dest_path, "w") as fp:
            fp.write(self.driver.page_source)
        original_size = self.driver.get_window_size()
        required_width = self.driver.execute_script('return document.body.parentNode.scrollWidth')
        required_height = self.driver.execute_script('return document.body.parentNode.scrollHeight')
        self.driver.set_window_size(required_width, required_height)
        # driver.save_screenshot(path)  # has scrollbar
        self.driver.find_element_by_tag_name('body').screenshot(url)  # avoids scrollbar
        self.driver.set_window_size(original_size['width'], original_size['height'])        
        return dest_path



import os

try:
    import projects.scraper.src.scraper as scraper
except ImportError:
    import os
    import sys

    sys.path.append(os.getcwd())

from qai.scraper.scrapers.scraper import Scraper
from qai.scraper.scrapers.stealthscraper import StealthScraper


def get_scraper(
    stealth: bool = False,
) -> Scraper:
    s = StealthScraper() if stealth else Scraper()
    return s



import logging
import os
import re
import sys
from dataclasses import dataclass
from typing import Any

import ftfy
from bs4 import BeautifulSoup
from markdownify import markdownify as md

from qai.scraper.processors.writer import Writer
from qai.core import MetaFile, MetaDir

@dataclass
class MarkdownWriter(Writer):
    file_name: str = None
    body_width: int = None

    def process(
        self, soup: BeautifulSoup, instance_info: dict[str, Any] = None, typ=1
    ) -> BeautifulSoup:
        group: MetaDir = self._append_and_modify_filename("md", instance_info=instance_info)
        if not self.out_folder and not self.file_name:
            raise ValueError("out_folder must be specified for MarkdownWriter")
        if not self.file_name:
            ii = instance_info
            self.file_name = ii["web_file"].path(self.out_folder)
        ## TODO check when this needs to happen
        ## switch extension to .md
        if not isinstance(self.file_name, str):
            self.file_name = str(self.file_name)
        if not self.file_name.endswith(".md"):
            f, __ = os.path.splitext(self.file_name)
            self.file_name = f + ".md"
        with open(self.file_name, "w") as fp:
            print(f"Writing markdown to {self.file_name}")
            html = ftfy.fix_text(str(soup.prettify()))
            if typ == 0:
                new_str = md(html)
            elif typ == 1:
                import html2text

                h = html2text.HTML2Text(bodywidth=self.body_width)
                new_str = h.handle(html)
            logging.debug(f"Writing markdown to {self.file_name}")
            fp.write(new_str) 
            if self.meta:
                mf = MetaFile(path=self.file_name, metadata={"instance_info": instance_info})
                if group:
                    group.add_file(mf)
                else:
                    self.meta.add_file(mf)


        return soup




from qai.core.meta import MetaBase


class ScrapeMeta(MetaBase):
    subfolder: str
    start_urls: list[str]
    dest_dir: str
    exclude_regexs: list[str]
    exclude_urls: list[str]
    exclude_parameters: bool
    max_depth: int



import glob
import os
import re
import traceback
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar

from bs4 import BeautifulSoup
from pi_conf import Config
from pi_log import getLogger
from qai.core import Meta, MetaFile

from qai.scraper.processors.processor_factory import get_processor
from qai.scraper.processors.processors import Processor
# from qai.scraper.scrapers.meta import WebObj
from qai.scraper.utils.bs_utils import make_soup as bs_util_make_soup
from qai.scraper.utils.bs_utils import make_soup_from_file

from .meta import DirectoryInfo, InstanceInfo

log = getLogger(__name__)

DEFAULT_PARSER = "lxml"


class MultiFilter:
    def __init__(
        self,
        configs: list[str] | list[dict[str, Any]],
        meta: Meta = None,
    ):
        # for config in configs:
        #     if "subfolder" not in config.get("default_options", {}):
        #         raise ValueError("subfolder must be specified in config.default_options for MultiFilter")
        self.configs = []
        for config in configs:
            if isinstance(config, str):
                config = Config.from_str(config)
            else:
                config = Config.from_dict(config)
            self.configs.append(config)
        self.meta = meta

    def process_directory(self, in_folder: str = None, in_group: str = None):
        for config in self.configs:
            f = Filter(config, meta=self.meta)
            f.process_directory(in_folder=in_folder, in_group=in_group)


class Filter:
    process_file_count: ClassVar[int] = 0

    def __init__(
        self,
        config: str | dict[str, Any],
        meta: Meta = None,
        # subfolder: str = "",
    ):
        if isinstance(config, str):
            self.config = Config.from_str(config)
        else:
            self.config = Config.from_dict(config)
        self.meta = meta

        self.processer_configs = get_processor_config_map(config)
        self.default_ops: dict[str, Any] = config.get("default_options", {})
        self.default_pipeline = self.default_ops.get("pipeline")
        self._proc_name2config_map = {}
        for proc_config in config.get("processors", []):
            proc_name = proc_config.get("name")
            self._proc_name2config_map[proc_name] = proc_config
        self.default_parser = config.get("default_parser", DEFAULT_PARSER)

    # def _attach_instance_info(self, config: dict[str, Any], key: str, value) -> dict[str, Any]:
    #     if not "_instance_info_" in config:
    #         config["_instance_info_"] = {}
    #     config["_instance_info_"][key] = value

    def process_file(
        self,
        filename_or_webfile_name: MetaFile | str | Path,
        directory_info: DirectoryInfo = None,
    ) -> list[BeautifulSoup]:
        list_cfgs = self._get_matching_configs(filename_or_webfile_name)
        if not list_cfgs:
            return []

        log.debug(f"filter:process_file: {filename_or_webfile_name}")
        id = self.process_file_count
        self.process_file_count += 1
        if isinstance(filename_or_webfile_name, Path):
            filename_or_webfile_name = str(filename_or_webfile_name)
        if isinstance(filename_or_webfile_name, str):
            # if self.meta:
            #     uri = self.meta.relative_to(filename_or_webfile_name)
            # else:
            #     uri = filename_or_webfile_name
            filename_or_webfile_name = MetaFile(
                path=filename_or_webfile_name,
                # source_uri=filename_or_webfile_name,
                # _meta=self.meta,
            )
        webfile: MetaFile = filename_or_webfile_name
        soups = []
        instance_info = InstanceInfo(
            # _uri=webfile._uri,
            web_file=webfile,
            _meta=self.meta,
            directory_info=directory_info,
        )
        instance_info._meta = self.meta

        for i, cfg in enumerate(list_cfgs):
            instance_info.config = cfg
            instance_info.config_index = i
            # soup = self.filter_html_file(webfile.full_path(), cfg, instance_info=instance_info)
            soup = self.filter_html_file(webfile.path, cfg, instance_info=instance_info)
            soups.append(soup)
        return soups

    def process_soup(self, soup: BeautifulSoup):
        return self.filter_soup(soup)

    def _should_include(self, file_info: str | MetaFile, config: dict[str, Any]) -> bool:
        """
        Returns True if the file should be included,
        False if it should be excluded.
        """
        if (
            not "url_regex" in config
            and not "file_regex" in config
            and not "exclude_url_regex" in config
            and not "exclude_file_regex" in config
        ):
            ### Not specifying anything matches all files
            return True
        @dataclass
        class NameSource:
            uri: str
            source_uri: str

        if isinstance(file_info, (str, Path)):
            bn = os.path.basename(file_info)
            file_info = NameSource(uri=bn, source_uri=str(file_info))
        else:
            file_info = NameSource(uri=file_info.path, source_uri="")
        
        for typ in ["source_uri", "uri"]:
            to_match = getattr(file_info, typ)
            typp = {"source_uri": "url", "uri": "file"}[typ]

            include = f"{typp}_regex"
            exclude = f"exclude_{typp}_regex"
            if exclude in config:
                if re.match(config[exclude], to_match):
                    log.debug(f"Excluded {to_match} to {config[exclude]}")
                    return False
                else:
                    log.debug(f"Did not exclude {to_match} to {config[exclude]}")
            if include in config:
                if re.match(config[include], to_match):
                    log.debug(f"Matched {to_match} to {config[include]}")
                    return True
                else:
                    log.debug(f"Did not match {to_match} to {config[include]}")

        return False

    def process_directory(
        self, in_group: str = None, in_folder: str = None, directory_info: DirectoryInfo = None
    ):
        """
        Needs in_group or in_folder to be specified
        """
        if not in_group and not in_folder:
            raise ValueError("in_group or in_folder must be specified")
        if directory_info is None:
            directory_info = DirectoryInfo(_meta=self.meta, in_folder=str(in_folder))

        if in_group is None and in_folder:
            for f in glob.glob(f"{in_folder}/**"):
                log.debug(f"Filter:Processing {str(f)}")
                self.process_file(f, directory_info=directory_info)
        else:
            directory_info.in_folder = in_folder
            g = self.meta.get_dir(in_group)
            files = list(self.meta.get_files(in_group))
            for f in files:
                print(f"Filter:Processing {str(f)}, len={len(files)}")
                log.debug(f"Filter:Processing {str(f)}, len={len(files)}")
                self.process_file(f, directory_info=directory_info)

    def _subsitute_values(self, config: dict, filename: MetaFile) -> dict:
        for k, v in config.items():
            if isinstance(v, str):
                try:
                    v = v.replace("<filename>", filename.path)
                    v = v.replace("<basename>", filename.path.name)
                    # v = v.replace("<group>", self.subfolder)
                except:
                    pass
                config[k] = v
        return config

    def _get_matching_configs(self, file_name: MetaFile) -> list[dict[str, Any]]:
        ## If we have a list of configs, then we need to match the file_name to the regex
        if "pipeline" in self.config:
            configs = []
            for url_config in self.config.get("pipeline", []):
                if not self._should_include(file_name, url_config):
                    continue

                ### Not specifying anything matches all files
                base_config = self.default_ops.copy()
                base_config.update(url_config)
                configs.append(self._subsitute_values(base_config, filename=file_name))
            return configs
        ## If no configs are specified, then we just use the default config for all files
        base_config = self.default_ops.copy()
        return [self._subsitute_values(base_config, filename=file_name)]

    def _get_pipeline(self, processor_names: list[str], config: dict[str, Any]) -> list[Processor]:

        pipeline = []
        for proc_name in processor_names:
            base_config = config.copy()
            pconfig = self._proc_name2config_map.get(proc_name, {}).copy()
            base_config.update(pconfig)
            proc = get_processor(name=proc_name, config=base_config, meta=self.meta)
            pipeline.append(proc)
            log.debug(f"Adding processor {proc.name}")
        return pipeline

    def _make_processors(self, config: dict[str, Any]) -> list[Processor]:
        proc_names = config.get("pipeline", self.default_pipeline)
        if not proc_names:
            raise ValueError(f"No pipeline found for inside config")

        ### TODO: Need to handle multiple pipelines later
        if len(proc_names) > 0 and isinstance(proc_names[0], dict):
            proc_names = proc_names[0]
            proc_names = proc_names.get("pipeline", self.default_pipeline)
        return self._get_pipeline(processor_names=proc_names, config=config)

    def filter_html_file(
        self, html_file_or_fp: str, config: dict[str, Any], instance_info: dict[str, Any] = None
    ) -> BeautifulSoup:
        soup = make_soup_from_file(html_file_or_fp)
        return self.filter_soup(soup, config=config, instance_info=instance_info)

    def filter_html_string(self, html: str, config: dict[str, Any] = None) -> BeautifulSoup:
        return self.filter_soup(self.make_soup(html), config)

    def filter_soup(
        self, soup: BeautifulSoup, config: dict[str, Any] = None, instance_info=None
    ) -> BeautifulSoup:
        if config is None:
            config = self.config
        procs = self._make_processors(config)

        return self._filter_soup(soup, procs, config=config, instance_info=instance_info)

    def _filter_soup(
        self,
        soup: BeautifulSoup,
        procs: list[Processor] = None,
        config: dict[str, Any] = None,
        instance_info: InstanceInfo = None,
    ) -> BeautifulSoup:
        tags = config.get("tags")
        if tags and self.meta:
            if isinstance(tags, str):
                tags = [tags]
        for i, p in enumerate(procs):
            log.debug(f"    {i}, {p.name} nodes={len(soup())}")
            soup = p.process(soup, instance_info=instance_info)

        return soup

    def filter_url(self, url: str):
        f = self.meta.get_file(url)
        return self.filter_html_string(f)

    def filter_urls(self, urls: list[str]):
        for url in urls:
            self.filter_url(url)

    @staticmethod
    def make_soup(html_str_or_path_or_fp: str, default_parser: str = "lxml") -> BeautifulSoup:
        if isinstance(html_str_or_path_or_fp, MetaFile):
            html_str_or_path_or_fp = str(html_str_or_path_or_fp)
        return bs_util_make_soup(html_str_or_path_or_fp, default_parser=default_parser)


def get_processor_config_map(config: dict[str, Any]) -> dict[str, Any]:
    cfg = config.copy()
    d = {}
    for p in cfg.get("processors", []):
        name = p.get("name")
        d[name] = p
    return d



import os

from pi_conf import load_config
from qai.scraper.filters.filter import Filter as Filter
from qai.scraper.scrapers.scraper import Scraper as Scraper

cfg = load_config("qrev-ai")
cfg.to_env(ignore_complications=True)

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = os.path.join(ROOT_DIR, "..", "data")
PARSE_CONFIG_DIR = os.path.join(ROOT_DIR, "..", "parse_configs")
try:
    import importlib.metadata

    VERSION = importlib.metadata.version("qai.scraper")
except Exception:
    VERSION = "0.0.0"



import io
import os
import re
from pathlib import Path
from typing import Any, Set, Union
from urllib.parse import urljoin, urlsplit

import validators
from bs4 import BeautifulSoup, element
from lxml.html.clean import Cleaner


def base_url(url):
    split_url = urlsplit(url)
    return f"{split_url.scheme}://{split_url.netloc}"


def url_with_path(url):
    split_url = urlsplit(url)
    url = f"{split_url.scheme}://{split_url.netloc}{split_url.path}"
    if url.endswith("/"):
        url = url[:-1]
    return url


def url_with_www(url):
    split_url = urlsplit(url)
    url = f"{split_url.scheme}://www.{split_url.netloc}{split_url.path}"
    if url.endswith("/"):
        url = url[:-1]
    return url


def validate_url(url):
    u = url_with_path(url)
    e = validators.url(url)
    if e == True:
        return e
    else:
        url = url_with_www(url)
        return validators.url(url)


def make_soup_from_file(file_or_fp: str | io.FileIO, default_parser: str = "lxml") -> BeautifulSoup:
    if isinstance(file_or_fp, io.IOBase):
        file_or_fp = file_or_fp.read()
    else:
        file_or_fp = str(file_or_fp)
        potential_path = os.path.expanduser(file_or_fp)
        if not os.path.exists(potential_path):
            raise FileNotFoundError(f"File {file_or_fp} not found")
        with open(potential_path, "r") as fp:
            file_or_fp = fp.read()
    return BeautifulSoup(file_or_fp, default_parser)


def make_soup(
    html_str_or_path_or_fp: str | io.FileIO, default_parser: str = "lxml"
) -> BeautifulSoup:
    if isinstance(html_str_or_path_or_fp, io.IOBase):
        html_str_or_path_or_fp = html_str_or_path_or_fp.read()
    else:
        if isinstance(html_str_or_path_or_fp, str):
            potential_path = os.path.expanduser(html_str_or_path_or_fp)
        else:
            potential_path = html_str_or_path_or_fp
        if os.path.exists(potential_path):
            with open(potential_path, "r") as fp:
                html_str_or_path_or_fp = fp.read()
    return BeautifulSoup(html_str_or_path_or_fp, default_parser)

def html_strip(html_str: str) -> str:
    s = str(html_str).replace("\n", " ")
    s = s.replace("\t", " ")
    s = re.sub(r"\s+", " ", s)
    ## Remove spaces before punctuation
    s = re.sub(r"\s+[.]", ".", s)
    s = re.sub(r"\s+[?]", "?", s)
    s = re.sub(r"\s+[!]", "!", s)
    s = re.sub(r"\s+[;]", ";", s)
    s = re.sub(r"\s+[,]", ",", s)
    ## Correctly format percentages (50%)
    s = re.sub(r"(\d+)\s+[%]", r"\1%", s)
    ## Times symbol (3x speed, or 4x4)
    s = re.sub(r"(\d+)\s+x(\s)", r"\1x\2", s)
    s = re.sub(r"(\d+)x\s+(\d+)", r"\1x\2", s)
    return s.strip()

def sanitize(dirty_html: str):
    cleaner = Cleaner(
        page_structure=True,
        meta=True,
        embedded=True,
        links=True,
        style=True,
        processing_instructions=True,
        inline_style=True,
        scripts=True,
        javascript=True,
        comments=True,
        frames=True,
        forms=True,
        annoying_tags=True,
        remove_unknown_tags=True,
        safe_attrs_only=False,
        safe_attrs=frozenset(["src", "color", "href", "title", "class", "name", "id"]),
        remove_tags=("span", "font", "div"),
    )

    return cleaner.clean_html(dirty_html)


def clean_soup(html_file_or_fp: str | Path) -> BeautifulSoup:
    from qai.scraper.processors.tag_remover import CleanHtml
    if isinstance(html_file_or_fp, Path):
        html_file_or_fp = str(html_file_or_fp)
    if isinstance(html_file_or_fp, str):
        with open(html_file_or_fp) as fp:
            soup = BeautifulSoup(fp, "html.parser")
    else:
        soup = BeautifulSoup(html_file_or_fp, "html.parser")
    soup = CleanHtml().process(soup)
    return soup


def get_tags_and_links(
    original_url, file_or_soup: str | BeautifulSoup, exclude_parameters=True, depth: int = 1
) -> list[tuple[element.Tag, str]]:
    soup = make_soup(file_or_soup)
    original_url = original_url.replace("www.", "")
    url_is_https = original_url.startswith("https")
    url_is_http = not url_is_https and original_url.startswith("http")
    url_without_http =  "/" + re.sub(r'^https?:\/\/', '', original_url)
    b_url = base_url(original_url)
    # print(f"{original_url} -> b_url=", b_url)
    found_links = {}

    for a in soup.find_all("a", href=True):
        link = a["href"]
        if link.startswith("http"): ## absolute url
            if not link.startswith(b_url): ## not our domain
                continue
        else: ## Relative url
            link = urljoin(original_url, link)
            # if link.startswith("/"): # absolute to root
            #     link = f"{b_url}{link}"
            # else: # relative to current page
            #     link = os.path.abspath(os.path.join(url_without_http, link))
            #     if url_is_http:
            #         link = f"http:/{link}"
            #     if url_is_https:
            #         link = f"https:/{link}"
        if exclude_parameters:
            link = url_with_path(link)

        if validate_url(link):
            if link not in found_links:
                found_links[link] = a

    return list([(v, k) for k, v in found_links.items()])

def get_links(
    original_url, file_or_soup: str | BeautifulSoup, exclude_parameters=True, depth: int = 1
) -> Set[str]:
    tags_and_links = get_tags_and_links(original_url, file_or_soup, exclude_parameters, depth)
    return list(set([link for _, link in tags_and_links]))



import logging
import os
import random
import time
from hashlib import sha256
from urllib.parse import urlsplit

import undetected_chromedriver as uc
from bs4 import BeautifulSoup

from .scraper import Scraper

## Can check at https://useragentstring.com/, https://useragentstring.com/pages/Chrome/
user_agents = [
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.79 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.84 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.77 Safari/537.36",
]


class StealthScraper(Scraper):
    def __init__(self, options=None, driver=None):
        self.time_between_requests = 4
        self.time_jitter_between_requests = 6
        self.last_request_time = 0
        self.window_size = (1200, 800)
        self.window_size_range = (600, 600)
        self.window_pos = (400, 400)
        self.window_pos_range = (300, 300)
        super().__init__(options, driver)

    def _new_driver(
        self,
        quit=True,
        options=None,
        driver=None,
        random_user_agent=False,
        version_main=None,  ## Specific version of chrome to use
    ):
        if quit:
            self.quit()
        if options is None:
            options = uc.ChromeOptions()
            options.add_argument("--disable-blink-features=AutomationControlled")
            options.add_argument("--headless=True")
        if driver is None:
            driver = uc.Chrome(use_subprocess=True, options=options, version_main=version_main)
            driver.execute_cdp_cmd(
                "Network.setUserAgentOverride",
                {"userAgent": random.choice(user_agents)},
            )
            driver.execute_script(
                "Object.defineProperty(navigator, 'webdriver', {get: () => undefined})"
            )
            driver.set_window_position(
                random.randint(
                    self.window_pos[0] - self.window_pos_range[0],
                    self.window_pos[0] + self.window_pos_range[0],
                ),
                random.randint(
                    self.window_pos[1] - self.window_pos_range[1],
                    self.window_pos[1] + self.window_pos_range[1],
                ),
            )
            driver.set_window_size(
                random.randint(
                    self.window_size[0] - self.window_size_range[0],
                    self.window_size[0] + self.window_size_range[0],
                ),
                random.randint(
                    self.window_size[1] - self.window_size_range[1],
                    self.window_size[1] + self.window_size_range[1],
                ),
            )
        self.driver = driver

    @staticmethod
    def base_url(url):
        split_url = urlsplit(url)
        return f"{split_url.scheme}://{split_url.netloc}"

    def get(self, url):
        self._wait_if_needed()
        self._new_driver()
        d = self.driver.get(url)
        self.last_request_time = time.time()
        return d

    def _wait_if_needed(self):
        random_wait = random.randint(
            self.time_between_requests,
            self.time_between_requests + self.time_jitter_between_requests,
        )
        dif = (self.last_request_time + random_wait) - time.time()
        if dif > 0:
            time.sleep(dif)

    def quit(self):
        self.driver.quit()
